let locale = "en_US";
let instanceProperties = {};
let ssoList = document.getElementById("available_sso");
let private_key = ""; // will be rsa key
let enterprise_data = ""; // is the response from the get_enterprise_data api call

let username;
let password;
let region;
let keeperRegions = ["us", "eu", "au"];
let sessionToken;
let data_key;
let waitForTwoFactor;
let deviceApprovalChannel;
let twoFactorChannel = "two_factor_channel_generic";
let twoFactorToken;
let twoFactorChannels;
let deviceApprovalChannels;

const WAIT_FOR_2FA_APPROVAL_SECONDS = 20;

const twoFactorChannelNames = {
    "TWO_FA_CT_NONE": "None",
    "TWO_FA_CT_TOTP": "Authenticator (TOTP)",
    "TWO_FA_CT_SMS": "Text message (SMS)",
    "TWO_FA_CT_DUO": "Duo Security",
    "TWO_FA_CT_RSA": "RSA SecurID",
    "TWO_FA_CT_BACKUP": "Backup code",
    "TWO_FA_CT_U2F": "U2F Security Key",
    "TWO_FA_CT_WEBAUTHN": "WebAuthn",
    "TWO_FA_CT_KEEPER": "Keeper Push",
    "TWO_FA_CT_DNA": "Keeper DNA device",
    "EMAIL": "Email",
    "KEEPER_PUSH": "Keeper Push",
    "TFA": "Two Factor Authentication",
    "KEEPER_DNA": "Keeper DNA"
};

function getDisplayNameForTwoFactorChannel(channel) {
    if (!channel) {
        return "Other";
    }
    let displayName = twoFactorChannelNames[channel.toUpperCase()];
    if (displayName === null) {
        return channel;
    }
    return displayName;
}

const twoFactorChannelMessages = {
    "two_factor_channel_duo": "Please enter the code from your Duo app.",
    "duo": "Please enter the code from your Duo app.",
    "two_factor_channel_email": "Your code was sent via email.",
    "email": "Your code was sent via email.",
    "two_factor_channel_generic": "Please select or enter a value.",
    "generic": "Please select or enter a value.",
    "two_factor_channel_google": "Please enter the code in your Authenticator (TOTP) app.",
    "google": "Please enter the code in your Authenticator (TOTP) app.",
    "totp": "Please enter the code in your Authenticator (TOTP) app.",
    "two_factor_channel_push": "Please enter the verification code on your DNA device.",
    "keeper_push": "Please enter the verification code on your DNA device.",
    "keeper_dna": "Please enter the verification code on your DNA device.",
    "push": "Please enter the verification code on your DNA device.",
    "two_factor_channel_rsa": "Please enter your RSA passcode.",
    "rsa": "Please enter your RSA passcode.",
    "two_factor_channel_sms": "Please enter the code sent via text message.",
    "sms": "Please enter the code sent via text message.",
    "two_factor_channel_u2f": "Please use your U2F Security Key to authenticate.",
    "u2f": "Please use your U2F Security Key to authenticate.",
    "two_factor_channel_voice": "You will receive a phone call with your code momentarily.",
    "voice": "You will receive a phone call with your code momentarily."
};

function getMessageForTwoFactorChannel(channel) {
    if (!channel) {
        return "Select an entry";
    }
    let message = twoFactorChannelMessages[channel.toLowerCase()];
    if (message === null) {
        let pos = channel.lastIndexOf("_");
        let suffix = channel.substring(pos + 1);  // returns whole string if not found
        message = twoFactorChannelMessages[suffix.toLowerCase()];
        if (message === null) {
            return channel;
        }
    }
    return message;
}


const SERVER_MONITOR_INTERVAL_SECS = 20;

const greenLightColor = "#3aac14";
const redLightColor = "#F7324A";
const normalTextColor = "black";

let isLoggedIn = false;

/**
 * While editing the configuration, the changed set of properties is stored here
 * and not committed until the user hits the SAVE button.
 *
 * The sks properties are sent as an object with key "sks_properties".
 * The value is an array containing an object for each property.
 * The fields of each property are:
 *     category
 *     context
 *     display_name
 *     editable
 *     edit_name
 *     long_name
 *     name
 *     value
 *
 * All values are strings except for "editable", which is a boolean value.
 */
let edited_sks_properties = [];
const DEFAULT_HSM_TYPE = "Mock Secure KeyStore for Testing";

/**
 * Requests the SSO Connect server to log in the user using the specified username and password.
 */
function login() {
    let twoFactorChannelUid;
    isLoggedIn = false;
    username = $("#username").val();
    password = $("#password").val();

    $.ajax({
        url: "/login",
        async: false,
        type: 'post',
        data: {
            region: $("#region").val(),
            username: username,
            password: password
        },
        dataType: 'json',
        success: function (_response, _textStatus, _jqXHR) {
            // region redirect?
            if (_response["region"]) {
                region = _response["region"];
                instanceProperties.keeper_region = region;
                setRegions();
            }

            let action = _response["action"];
            // Check result
            if (_response["result"] === "success") {
                if (action === "logged_in") {
                    isLoggedIn = true;
                    document.getElementById("password").value = '';
                    document.getElementById("two_factor_code").value = '';
                    getInstanceProperties(status);
                } else if (action === "require_device_approval") {
                    deviceApprovalChannel = _response["channel"];
                    document.querySelector("login-body").style.display = 'none';
                    document.querySelector("device-approval-channel-dialog").style.display = 'none';
                    document.querySelector("device-approval-dialog").style.display = 'block';
                    $("device-approval-dialog > div > p").text(getMessageForTwoFactorChannel(twoFactorChannel));
                    document.getElementById("device_approval_code").value = '';
                    document.getElementById("device-approval-error").style.display = 'none';
                } else if (action === "require_device_approval_channel") {
                    deviceApprovalChannels = _response["channel"];
                    if (deviceApprovalChannels) {
                        let deviceApprovalChannelSelector = document.getElementById('device_approval_channel');
                        deviceApprovalChannelSelector.options = [];
                        deviceApprovalChannels.forEach(function (item, index) {
                            deviceApprovalChannelSelector.options[deviceApprovalChannelSelector.options.length] = new Option(getDisplayNameForTwoFactorChannel(item), index);
                        });
                    }
                    document.querySelector("login-body").style.display = 'none';
                    document.querySelector("device-approval-dialog").style.display = 'none';
                    document.querySelector("device-approval-channel-dialog").style.display = 'block';
                    // $("device-approval-channel-dialog > div > p").text(getMessageForTwoFactorChannel(twoFactorChannel));
                    document.getElementById("device-approval-channel-error").style.display = 'none';
                } else if (action === "require_2fa") {
                    twoFactorChannelUid = _response["channel_uid"];
                    setup2faDialog(_response["channel"]);
                } else if (action === "require_2fa_channel") {
                    twoFactorChannels = _response["channel"];
                    if (twoFactorChannels) {
                        let twoFactorChannelSelector = document.getElementById('two_factor_channel');
                        twoFactorChannelSelector.options.length = 0;
                        twoFactorChannels.forEach(function (item, index) {
                            twoFactorChannelSelector.options[twoFactorChannelSelector.options.length] = new Option(getDisplayNameForTwoFactorChannel(item), index);
                        });
                    }
                    document.querySelector("login-body").style.display = 'none';
                    document.querySelector("device-approval-dialog").style.display = 'none';
                    document.querySelector("device-approval-channel-dialog").style.display = 'none';
                    document.querySelector("two-factor-dialog").style.display = 'none';
                    document.querySelector("two-factor-channel-dialog").style.display = 'block';
                    $("two-factor-channel-dialog > div > p").text(getMessageForTwoFactorChannel(twoFactorChannel));
                    document.getElementById("two-factor-channel-error").style.display = 'none';
                } else {
                    alert("Unknown action from login: " + action);
                }
            } else {

                alert("Unable to login.  Please try again.\n\n" + _response["message"]);
                logout(true);
                password = $("#password").val();  // retain the password for another try
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            alert("Unable to login.  Please try again.\n\n" + _errorThrown);
        }
    });
}

/**
 * Sets up the display for regular or Duo 2FA.
 * This is called when the "require_2fa" action is received from the server.
 */
function setup2faDialog(channel, action) {
    twoFactorChannel = channel;

    document.querySelector("login-body").style.display = 'none';
    document.querySelector("device-approval-dialog").style.display = 'none';
    document.querySelector("two-factor-channel-dialog").style.display = 'none';
    document.getElementById("two_factor_code").value = '';
    document.getElementById("two-factor-error").style.display = 'none';

    // Duo requires a different interface    
    if ((channel.toLowerCase().indexOf("duo") >= 0) && action && (action !== "require_2fa_for_duo")) {
        document.querySelector("two-factor-dialog").style.display = 'none';
        document.getElementById("duo_select_dialog").style.display = 'block';
        return;
    }

    // Non-duo 2FA
    document.querySelector("two-factor-dialog").style.display = 'block';
    $("two-factor-dialog > div > p").text(getMessageForTwoFactorChannel(channel));
}

/**
 * Called when the 2FA method is Duo and the user selects a sub-channel.
 * This is similar to selectTwoFactorChannel().
 *
 * @param subChannel is either "push", "sms", or "voice"
 */
function select2faDuo(subChannel) {
    document.getElementById("duo_select_dialog").style.display = 'none';
    selectTwoFactorChannel(null, subChannel);
}

/**
 * Changes some settings here and then asks the SSO Connect server to log the user out.
 *
 */
function logout(noserver) {
    username = "";
    password = "";
    sessionToken = "";
    data_key = "";
    enterprise_data = "";
    waitForTwoFactor = undefined;
    data_key = undefined;

    document.getElementById("logged_in").style.display = 'none';
    document.getElementById("status").style.display = 'none';
    document.getElementById("configure").style.display = 'none';
    document.getElementById("initialize").style.display = 'none';
    document.querySelector("login-body").style.display = 'block';
    document.querySelector("two-factor-dialog").style.display = 'none';
    document.querySelector("two-factor-channel-dialog").style.display = 'none';
    document.querySelector("device-approval-dialog").style.display = 'none';
    document.querySelector("device-approval-channel-dialog").style.display = 'none';
    document.getElementById("duo_select_dialog").style.display = 'none';
    document.getElementById("login").style.display = 'block';

    if (!isLoggedIn) {
        return;
    }

    isLoggedIn = false;
    if (!noserver) {
        $.ajax({
            url: "/logout",
            async: true,
            type: 'GET',
            contentType: "text/plain; charset=UTF-8",
            data: '',
            dataType: 'json',
            success: function (_response, _textStatus, _jqXHR) {
            },
            error: function (_jqXHR, _textStatus, _errorThrown) {
            }
        });
    }
}

/**
 * Retrieves the properties (settings) from the SSO Connect server.
 * Update the display if successful.
 */
function getInstanceProperties(callback, errorCallBack) {
    $.ajax({
        url: "/properties/",
        async: true,
        type: 'GET',
        contentType: "text/plain; charset=UTF-8",
        data: '',
        dataType: 'json',
        success: function (response, _textStatus, _jqXHR) {
            instanceProperties = response;
            let ver = "v " + instanceProperties.version.substring(1);
            if (instanceProperties.build && instanceProperties.environment) {
                ver = ver + "." + instanceProperties.build;
            }
            document.getElementById("version").textContent = ver;
            document.getElementById("version-sidebar").textContent = ver;
            if (instanceProperties.idp_type) {
                document.getElementById("idp_type").value = instanceProperties.idp_type;
            }
            setRegions();
            if (instanceProperties.result_code === "re_login") {
                logout();
            } else if (callback) {
                callback();
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            if (errorCallBack) {
                errorCallBack();
            } else {
                alert("Error: unable to get instance properties; the SSO Connect server is not responding.");
                instanceProperties.status = "Error";
                status();
            }
        }
    });
}

/**
 * Set the values in the region pull-down list on the login screen.
 * Requires the instance properties "keeper_regions" and "keeper_region".
 */
function setRegions() {
    let regionList = instanceProperties.keeper_regions;
    if (regionList) {
        keeperRegions = regionList;
    }

    // Set the UI elements
    let regions = document.getElementById("region");
    regions.options = [];
    keeperRegions.forEach((regionStr, index) => {
        regions.options[index] = new Option(getRegionName(regionStr), regionStr);
    });
    document.getElementById("region").value = instanceProperties.keeper_region;
}

/**
 * Checks the status of the Keeperapp connection.  May force a re-login.
 */
function checkKeeperappStatus() {

    $.ajax({
        url: "/keeperappstatus/",
        async: true,
        type: 'GET',
        contentType: "text/plain; charset=UTF-8",
        data: '',
        dataType: 'json',
        success: function (response, _textStatus, _jqXHR) {
            let keeperappStatus = response.keeperapp_connection;

            if (response.result_code === "re_login") {
                logout();
            } else {
                if (!keeperappStatus) {
                    startConnectionErrorHandler();
                } else {
                    stopConnectionErrorHandler();
                }
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            console.log("Error: the SSO Connect server is not responding.");
            instanceProperties.status = "Error";
            logout();
        }
    });
}

/**
 * Retrieves the list of SSO service providers.
 * Called after login if the SSO Connect node hasn't been initialized yet.
 */
function getEnterpriseData() {
    $.ajax({
        url: "/get_sso_service_providers",
        async: true,
        type: 'get',
        contentType: "text/plain; charset=UTF-8",
        dataType: 'json',
        success: function (response, _textStatus, _jqXHR) {
            if ("success" === response.result) {
                enterprise_data = response;
                ssoList.options.length = 0;
                if (enterprise_data.sso_services) {
                    enterprise_data.sso_services.forEach(function (item, index) {
                        if (!instanceProperties.sso_service_provider_id || instanceProperties.sso_service_provider_id === item.sso_service_provider_id) {
                            ssoList.options[ssoList.options.length] = new Option(item.name, index);
                        }
                    });
                }
            } else {
                alert("Error getEnterpriseData: " + response.message);
                logout();
            }
        },
        error: function (_jqXHR, textStatus, _errorThrown) {
            alert("Error getEnterpriseData: " + textStatus);
            logout();
        }
    });
}

/**
 * This calls the SSO Connect server which forwards the request to Keeper and returns the result.
 */
function initialize() {
    if (ssoList.length < 1) {
        return;
    }
    toggleBusy("initialize_button");

    $.ajax({
        url: "/initialize_sso_service_provider_instance",
        async: false, // Todo: This needs to become asynchronous at some point.
        type: 'POST',
        data: {
            "username": username,
            "locale": locale,
            "sso_service_provider_id": enterprise_data.sso_services[ssoList.options[ssoList.selectedIndex].value].sso_service_provider_id,
            "sso_service_provider_name": enterprise_data.sso_services[ssoList.options[ssoList.selectedIndex].value].name,
            "sso_instance_id": instanceProperties.sso_instance_id
        },
        dataType: 'json',
        success: function (response, _textStatus, _jqXHR) {
            if ("success" === response.result) {
                initializeBusyStatus();
            } else {
                alert("Error: " + response.message);
                toggleBusy("initialize_button", "Connect");
                logout();
            }
        },
        error: function (_jqXHR, textStatus, _errorThrown) {
            alert("Error: " + textStatus);
            toggleBusy("initialize_button", "Connect");
        }
    });
}

/**
 * This pings the server to see whether the SSO Connect instance is initialized.
 */
function initializeBusyStatus() {
    $.ajax({
        url: "/ping/",
        async: true,
        type: 'GET',
        contentType: "text/plain; charset=UTF-8",
        success: function (response, _textStatus, _jqXHR) {
            if (response.status === "Ready") {
                toggleBusy("initialize_button", "Connect");
                setTimeout(completeInitialization, 2000);
            } else {
                setTimeout(initializeBusyStatus, 1000);
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            setTimeout(initializeBusyStatus, 1000);
        }
    });
}

function toggleBusy(element, buttonText) {
    let e = document.getElementById(element);
    if (buttonText) {
        e.innerHTML = buttonText;
        e.style.pointerEvents = 'auto';
    } else {
        e.innerHTML = "&nbsp;<img src='/static/auth_loader.gif' alt='loading'>&nbsp;"
        e.style.pointerEvents = 'none';
    }
}

/**
 * This function is called when a user clicks on "prefer post" or "prefer redirect" for SLO or SSO.
 *
 * @param endpoint  'slo' or 'sso'
 * @param setting 'post' or 'redirect'
 * @returns void
 */
function setEndpointPreference(endpoint, setting) {

    let on = "inline";
    let off = "none";

    document.getElementById(endpoint + "PreferPost").style.display = (setting === "post") ? on : off;
    document.getElementById(endpoint + "ChoosePost").style.display = (setting === "post") ? off : on;
    document.getElementById(endpoint + "PreferRedirect").style.display = (setting === "post") ? off : on;
    document.getElementById(endpoint + "ChooseRedirect").style.display = (setting === "post") ? on : off;

    // Save the settings here because we aren't on the config page.
    saveSettings();
}

function completeInitialization() {
    document.getElementById("login").style.display = 'none';
    getInstanceProperties(status);
}

function status() {
    edited_sks_properties = [];

    document.getElementById("menu_item_status").classList.add('active');
    document.getElementById("menu_item_config").classList.remove('active');

    let baseUrl = "https://" + instanceProperties.sso_connect_host + ((instanceProperties.sso_ssl_port === "443") ? "" : (":" + instanceProperties.sso_ssl_port)) + "/sso-connect";
    if (instanceProperties.domain) {
        document.getElementById("domain").textContent = instanceProperties.domain;
    }
    if (instanceProperties.invite_new_users) {
        document.getElementById("invite_new_users").textContent = instanceProperties.invite_new_users.charAt(0).toUpperCase() + instanceProperties.invite_new_users.slice(1);
    } else {
        document.getElementById("invite_new_users").textContent = "False";
    }
    if (instanceProperties.sign_idp_traffic) {
        document.getElementById("sign_value").textContent = instanceProperties.sign_idp_traffic.charAt(0).toUpperCase() + instanceProperties.sign_idp_traffic.slice(1);
    } else {
        document.getElementById("sign_value").textContent = "False";
    }

    // Set the SSOConnect green/red Status
    let running = document.getElementById("running");
    running.textContent = instanceProperties.status;
    if ("Running" === instanceProperties.status) {
        running.style.color = greenLightColor;
        running.textContent = "Running";
        document.getElementById("running_dot").style.background = greenLightColor;
    } else {
        running.style.color = redLightColor;
        running.textContent = "Stopped";
        document.getElementById("running_dot").style.background = redLightColor;
    }

    // Network monitor of Keeperapp
    if (!instanceProperties.keeperapp_connection) {
        startConnectionErrorHandler();
    } else {
        stopConnectionErrorHandler();
    }

    if (!instanceProperties.sso_connect_host || !instanceProperties.sso_ssl_port) {
        document.querySelector("service-provider-head").style.display = 'none';
        document.querySelector("service-provider-configured").style.display = 'none';
        document.querySelector("service-provider-not-configured").style.display = 'block';
    } else {
        document.querySelector("service-provider-head").style.display = 'block';
        document.querySelector("service-provider-configured").style.display = 'block';
        document.querySelector("service-provider-not-configured").style.display = 'none';
        document.getElementById("sp_entity_value").textContent = baseUrl;
        if (instanceProperties.valid_host) {
            document.getElementById("sp_entity_value").style.color = greenLightColor;
            document.getElementById("acs_value").href = baseUrl + "/saml/sso";
            document.getElementById("sls_value").href = baseUrl + "/saml/slo";
            document.getElementById("client_login_value").href = baseUrl + "/saml/login?embedded";
            document.getElementById("client_logout_value").href = baseUrl + "/saml/logout";
        } else {
            document.getElementById("sp_entity_value").style.color = redLightColor;
            document.getElementById("acs_value").removeAttribute("href");
            document.getElementById("sls_value").removeAttribute("href");
            document.getElementById("client_login_value").removeAttribute("href");
            document.getElementById("client_logout_value").removeAttribute("href");
        }
        document.getElementById("acs_value").textContent = "saml/sso";
        document.getElementById("sls_value").textContent = "saml/slo";
        document.getElementById("client_login_value").textContent = "saml/login";
        document.getElementById("client_logout_value").textContent = "saml/logout";

        // SSL Certificate
        if ("Not Configured" === instanceProperties.ssl_expiration) {
            document.getElementById("sp_certificate_value").style.color = redLightColor;
            document.getElementById("sp_certificate_table").style.display = "none";
            document.getElementById("sp_certificate_message").style.display = "block";
        } else {
            document.getElementById("sp_certificate_value").style.color = normalTextColor;
            document.getElementById("sp_certificate_message").style.display = "none";
            document.getElementById("sp_certificate_table").style.display = "block";
            document.getElementById("sp_certificate_subject").textContent = (instanceProperties.ssl_subject ? instanceProperties.ssl_subject.substring(instanceProperties.ssl_subject.indexOf("=") + 1) : "");
            document.getElementById("sp_certificate_issuer").textContent = (instanceProperties.ssl_issuer ? instanceProperties.ssl_issuer.substring(instanceProperties.ssl_issuer.indexOf("=") + 1) : "");
            document.getElementById("sp_certificate_valid").textContent = instanceProperties.ssl_expiration;
            document.getElementById("sp_certificate_valid").style.color = (instanceProperties.ssl_expires_soon ? redLightColor : normalTextColor);

            let keyType = instanceProperties.key_type.toUpperCase();
            if (keyType === "EC") keyType = "Elliptic Curve";
            document.getElementById("sp_certificate_key_type").textContent = keyType;
            /*
            if (keyType === "RSA") {
                $("#key_type_RSA").prop('checked', true);
            } else {
                $("#key_type_EC").prop('checked', true);
            }
            */
        }
    }

    if (instanceProperties.idp) {

        document.getElementById("idp_value").textContent = instanceProperties.idp_entity;
        document.getElementById("sso_value").textContent = instanceProperties.idp_sso_label;
        document.getElementById("sso_value").href = instanceProperties.idp_sso;
        document.getElementById("slo_value").textContent = instanceProperties.idp_slo_label;
        document.getElementById("slo_value").href = instanceProperties.idp_slo;
        document.getElementById("sso_post_value").textContent = instanceProperties.idp_sso_post_label;
        document.getElementById("sso_post_value").href = instanceProperties.idp_sso_post;
        document.getElementById("slo_post_value").textContent = instanceProperties.idp_slo_post_label;
        document.getElementById("slo_post_value").href = instanceProperties.idp_slo_post;
        document.getElementById("certificate_value").textContent = "Certificate Expires " + instanceProperties.idp_cert_expires;
        document.getElementById("certificate_value").style.color = (instanceProperties.idp_cert_expires_soon ? redLightColor : normalTextColor);

        document.querySelector("identity-provider-configured").style.display = 'block';
        document.querySelector("identity-provider-not-configured").style.display = 'none';

        // KSC-269: Allow user to prefer Post or Redirect for SLO and SSO
        let on = "inline";
        let off = "none";
        let value = "post";

        if (instanceProperties.slo_endpoint_preference) {
            value = instanceProperties.slo_endpoint_preference;
            document.getElementById("sloPreferPost").style.display = (value === "post") ? on : off;
            document.getElementById("sloChoosePost").style.display = (value === "post") ? off : on;
            document.getElementById("sloPreferRedirect").style.display = (value === "post") ? off : on;
            document.getElementById("sloChooseRedirect").style.display = (value === "post") ? on : off;
        }

        if (instanceProperties.sso_endpoint_preference) {
            value = instanceProperties.sso_endpoint_preference;
            document.getElementById("ssoPreferPost").style.display = (value === "post") ? on : off;
            document.getElementById("ssoChoosePost").style.display = (value === "post") ? off : on;
            document.getElementById("ssoPreferRedirect").style.display = (value === "post") ? off : on;
            document.getElementById("ssoChooseRedirect").style.display = (value === "post") ? on : off;
        }

    } else {
        document.querySelector("identity-provider-configured").style.display = 'none';
        document.querySelector("identity-provider-not-configured").style.display = 'block';
        document.getElementById("sloPreferPost").style.display = "none";
        document.getElementById("sloChoosePost").style.display = "none";
        document.getElementById("sloPreferRedirect").style.display = "none";
        document.getElementById("sloChooseRedirect").style.display = "none";
        document.getElementById("ssoPreferPost").style.display = "none";
        document.getElementById("ssoChoosePost").style.display = "none";
        document.getElementById("ssoPreferRedirect").style.display = "none";
        document.getElementById("ssoChooseRedirect").style.display = "none";
    }

    document.getElementById("first_name_value").textContent = instanceProperties.map_first_name;
    document.getElementById("last_name_value").textContent = instanceProperties.map_last_name;
    document.getElementById("username_value").textContent = instanceProperties.map_email;
    document.getElementById("configure").style.display = 'none';

    if (instanceProperties.registered) {
        document.getElementById("status").style.display = 'block';
        document.getElementById("logged_in").style.display = 'flex';
    }

    if (instanceProperties.available_version) {
        document.getElementById("version_up_to_date").style.display = 'none';
        document.getElementById("available_version").style.display = 'inline-table';
        document.getElementById("available_version_text").textContent = "Version " + instanceProperties.available_version + " is available";
        document.getElementById("available_version_button").setAttribute("onclick", "window.location.href='" + instanceProperties.link + "';");
    } else {
        document.getElementById("available_version").style.display = 'none';
        document.getElementById("version_up_to_date").style.display = 'inline-table';
    }

    update_sks_gui();

    // Moved from another place because it assumed that getInstanceProperties had returned, but it hadn't yet
    if (instanceProperties.registered) {
        document.getElementById("login").style.display = 'none';
    } else {
        getEnterpriseData();
        document.querySelector("login-body").style.display = 'none';
        document.querySelector("two-factor-dialog").style.display = 'none';
        document.querySelector("two-factor-channel-dialog").style.display = 'none';
        document.querySelector("device-approval-dialog").style.display = 'none';
        document.querySelector("device-approval-channel-dialog").style.display = 'none';
        document.getElementById("initialize").style.display = 'block';
    }
}

function update_sks_gui() {
    // console.log("update_sks_gui called");

    // Update the SKS-related properties
    loadSKSProperties();

    // Set the SKS green/red Status =  Enabled or Disabled
    let running = document.getElementById("sks_running");
    let config_running_checkbox = document.getElementById("sks_config_running_checkbox");
    let config_running = document.getElementById("sks_config_running");

    if (getSKSstate()) {
        running.style.color = greenLightColor;
        running.textContent = getSKSstateString();
        config_running.style.color = greenLightColor;
        config_running.textContent = getSKSstateString();
        $(config_running_checkbox).prop("checked", true);
        document.getElementById("sks_running_dot").style.background = greenLightColor;
    } else {
        running.style.color = redLightColor;
        running.textContent = getSKSstateString();
        config_running.style.color = redLightColor;
        config_running.textContent = getSKSstateString();
        $(config_running_checkbox).prop("checked", false);
        document.getElementById("sks_running_dot").style.background = redLightColor;
    }

    // Set the SKS provider
    let sksTypeProperty = instanceProperties["sks_properties"].find(prop => prop.long_name === "secure_key_storage.type");
    let sksTypeValue = (sksTypeProperty ? sksTypeProperty.value : DEFAULT_HSM_TYPE);
    document.getElementById("sks_type").textContent = sksTypeValue;
    $("#config_sks_type").val(sksTypeValue);

    update_sks_properties_table();
}

function update_sks_properties_table() {
    let sksPropTableBody = $("#sks_properties tbody");

    // Remove the existing rows
    $(sksPropTableBody).empty();

    // Add a row for each property that is not the type property
    let rowNum = 1;
    for (let prop of edited_sks_properties) {
        if (prop.editable) {
            displaySKSProperty(sksPropTableBody, rowNum, prop);
            rowNum++;
        }
    }
}

/**
 * Returns true if the property was displayed.
 *
 * @param sksPropTableBody the HTML element to put items into
 * @param rowNum the row number
 * @param prop the property object
 * @returns true if the property was displayed in the SKS table, else false
 */
function displaySKSProperty(sksPropTableBody, rowNum, prop) {

    let val = prop.value;

    let rowEditIcon = $('<span class="ui-icon ui-icon-pencil floatLeft smaller-button"></span>');
    let rowTrashIcon = $('<span class="ui-icon ui-icon-trash floatLeft smaller-button"></span>');

    // Add the new row
    let rowHtml = '<tr><td>' + prop.edit_name + '</td><td>' + val + '</td><td class="cursorPointer alignCenter flexer"></td></tr>';
    let newRow = $(rowHtml);
    $(sksPropTableBody).append(newRow);

    // Add the icons to the third column
    let newMenuArea = $(newRow).children("td")[2];
    $(newMenuArea).append(rowEditIcon);
    $(newMenuArea).append(rowTrashIcon);

    // Add the clickListener for the icons for this row
    $(rowEditIcon).on("click", function (e) {
        editProperty(rowNum, prop.edit_name, prop.value);
    });
    $(rowTrashIcon).on("click", function (e) {
        deleteProperty(rowNum, prop.edit_name, prop.value);
    });

    return true;
}

function editProperty(rowNum, prop_edit_name, value) {
    let newValue = prompt('Enter a new value for ' + prop_edit_name, value);

    if (newValue !== null) {
        edited_sks_properties.map(prop => {
            if (prop.edit_name === prop_edit_name) {
                prop.value = newValue;
            }
        });
    }

    update_sks_properties_table();
}

function deleteProperty(rowNum, prop_edit_name, value) {
    // We can get duplicate entries.  Delete only the first one
    let found = false;
    let findFn = function (prop) {
        if (found) {
            return true;
        }
        if (prop.edit_name === prop_edit_name) {
            found = true;
            return false;
        }
        return true;
    };

    edited_sks_properties = edited_sks_properties.filter(findFn);

    update_sks_properties_table();
}

let connectionErrorStart = Date.now();
let connectionErrorTimer = null;

/**
 * This is called when the web app notices (via the SSO Connect server) that the
 * connection to Keeperapp is lost.
 *
 * If it has already been notified, it doesn't do anything.
 *
 * @returns nothing
 */
function startConnectionErrorHandler() {
    if (connectionErrorTimer === null) {
        console.log("Keeper Security server connection lost at " + (new Date()).toISOString());
        document.getElementById("connectionLost").style.display = 'block';
        connectionErrorStart = Date.now();

        connectionErrorTimer = setInterval((start) => {
            let now = Date.now();
            let elapsedSec = Math.floor((now - start) / 1000);
            let elapsedMin = Math.floor(elapsedSec / 60);
            elapsedSec = elapsedSec % 60;

            let timeString = ("0" + elapsedMin).slice(-2) + ":" + ("0" + elapsedSec).slice(-2);
            document.getElementById("connectionErrorTimer").textContent = timeString;
        }, 1000, connectionErrorStart);
    }
}

/**
 * This is called when the web app notices (via the SSO Connect server) that the
 * connection is up.  It hides the error message and refreshes the status page.
 *
 * If it has already been notified, it doesn't do anything.
 *
 * @returns nothing
 */
function stopConnectionErrorHandler() {
    if (connectionErrorTimer !== null) {
        console.log("Keeper Security server connection resumed at " + (new Date()).toISOString());
        clearInterval(connectionErrorTimer);
        connectionErrorTimer = null;
        document.getElementById("connectionLost").style.display = 'none';
        // Refresh the status
        getInstanceProperties(status);
    }
}

/**
 * Adds a new property to the SKS properties list.
 */
function addSKSProperty() {
    let category = "secure_key_storage";
    let context = $("#config_sks_type").val();
    context = instanceProperties["hsm_info"][context].name;
    let newProperty = prompt("Enter property name (short name)");
    if (newProperty) {
        let newValue = prompt("Enter a new value for " + newProperty);
        if (newValue) {
            let longName = category + "." + (context ? (context + ".") : "") + newProperty;

            // Add the new property to the properties list
            edited_sks_properties.push({
                category: category,
                context: context,
                display_name: newProperty,
                editable: true,
                edit_name: longName,
                long_name: longName,
                name: newProperty,
                value: newValue
            });

            update_sks_properties_table();
        }
    }
}

/**
 * Returns true if the property is an SKS property.
 * @param prop
 * @returns
 */
function isSKSProperty(prop) {
    return prop.startsWith("sks") || prop.startsWith("secure_key_storage");
}

/**
 * Copies from instanceProperties into edited_sks_properties.
 * Use this when beginning to edit the configuration, or when edits need to be reset.
 */
function loadSKSProperties() {
    // console.log("loadSKSProperties called");
    edited_sks_properties = JSON.parse(JSON.stringify(instanceProperties["sks_properties"]));

    specializeSKSProperties();
}

/**
 * Call this when the SAVE button is pushed on the configuration edit screen.
 * It copies the edited_sks_properties back into instanceProperties.
 *
 * @returns nothing
 */
function saveSKSProperties() {
    // console.log("saveSKSProperties called");
    instanceProperties["sks_properties"] = JSON.parse(JSON.stringify(edited_sks_properties));
}


/**
 * This undoes all changes to the SKS properties.
 * Hitting the main Cancel button also does the same thing.
 *
 * @returns nothing
 */
function resetSKSProperties() {
    loadSKSProperties();

    update_sks_properties_table();
}

function clearSKSProperties() {
    if (confirm("Do you really want to delete all of the SKS properties?")) {
        edited_sks_properties = edit_sks_properties.filter(prop => !prop.editable);

        update_sks_properties_table();
    }
}

/**
 * Returns true or false based on the state.
 * @returns
 */
function getSKSstate() {
    let props = instanceProperties["sks_properties"];
    if (edited_sks_properties && (edited_sks_properties.length > 0)) {
        props = edited_sks_properties;
    }

    return (String(props.filter(prop => prop.long_name === "secure_key_storage.enabled")[0].value) === "true");
}

function getSKSstateString() {
    return getSKSstate() ? "Enabled" : "Disabled";
}

function enableSKS(state) {
    let val = (String(state) === "true");
    let props = instanceProperties["sks_properties"];
    if (edited_sks_properties && (edited_sks_properties.length > 0)) {
        props = edited_sks_properties;
        // console.log("enableSKS: using edited props");
    }
    // console.log("enableSKS: setting enabled to " + val);


    props.filter(prop => prop.long_name === "secure_key_storage.enabled")[0].value = val;
    document.getElementById("sks_config_running").textContent = getSKSstateString();
    document.getElementById("sks_config_running").style.color = (val ? greenLightColor : redLightColor);

    // Ensure that secure_key_storage.type is set if SKS is turned on.
    if (val) {
        changeSKSType();
    }

    return val;
}

function changeSKSType() {
    let newType = $("#config_sks_type").val();
    let found = false;
    edited_sks_properties.map(prop => {
        if (prop.long_name === "secure_key_storage.type") {
            prop.value = newType;
            found = true;
        }
    });

    if (!found) {
        edited_sks_properties.push({
            category: "secure_key_storage",
            context: null,
            display_name: "type",
            editable: false,
            edit_name: "secure_key_storage.type",
            long_name: "secure_key_storage.type",
            name: "type",
            value: newType
        });
    }

    specializeSKSProperties();

    update_sks_properties_table();
}

function specializeSKSProperties() {

    // Add in the hsm-required properties
    let sksType = $("#config_sks_type").val();
    let hsmInfo = instanceProperties.hsm_info[sksType];
    let category = "secure_key_storage";
    let context = hsmInfo.name;

    // We don't want to have duplicate properties so make a list of the known properties
    let knownProperties = instanceProperties.sks_properties.map(prop => prop.long_name);

    Object.keys(hsmInfo.required_properties).forEach(function (key) {
        let longName = (category ? (category + ".") : "") + (context ? (context + ".") : "") + key;

        if (!knownProperties.includes(longName)) {
            // Add the new property to the properties list
            edited_sks_properties.push({
                category: category,
                context: context,
                display_name: key,
                editable: true,
                edit_name: longName,
                long_name: longName,
                name: key,
                value: hsmInfo.required_properties[key]
            });
        }
    });
    update_sks_properties_table();
}

function editConfiguration() {
    loadSKSProperties();  // resets edited_sks_properties

    document.getElementById("menu_item_status").classList.remove('active');
    document.getElementById("menu_item_config").classList.add('active');

    if (instanceProperties.sso_connect_host) {
        document.getElementById("sso_connect_host").value = instanceProperties.sso_connect_host;
    } else {
        document.getElementById("sso_connect_host").value = "";
    }
    if (instanceProperties.private_ip) {
        document.getElementById("private_ip").value = instanceProperties.private_ip;
    } else {
        document.getElementById("private_ip").value = "";
    }
    if (instanceProperties.private_port) {
        document.getElementById("private_port").value = instanceProperties.private_port;
    } else {
        document.getElementById("private_port").value = "";
    }
    document.getElementById("sso_ssl_port").value = instanceProperties.sso_ssl_port;
    document.getElementById("sign_idp_traffic").checked = (instanceProperties.sign_idp_traffic === "true");

    if (instanceProperties.map_first_name) {
        document.getElementById("map_first_name").value = instanceProperties.map_first_name;
    }
    if (instanceProperties.map_last_name) {
        document.getElementById("map_last_name").value = instanceProperties.map_last_name;
    }
    if (instanceProperties.map_email) {
        document.getElementById("map_email").value = instanceProperties.map_email;
    }

    document.getElementById('save_bottom').style.pointerEvents = 'auto';
    document.getElementById('save_top').style.pointerEvents = 'auto';
    document.getElementById('cancel_bottom').style.pointerEvents = 'auto';
    document.getElementById('cancel_top').style.pointerEvents = 'auto';

    document.getElementById("status").style.display = 'none';
    document.getElementById("configure").style.display = 'block';
}

function exportMetaData() {
    document.getElementById("export_sso_connect_metadata").click();
}

/**
 * This calls the SSO Connect server to cause it to restart.
 */
function restart() {
    $.ajax({
        url: "/restart",
        async: true,
        type: 'POST',
        contentType: "text/plain; charset=UTF-8",
        success: function (_response, _textStatus, _jqXHR) {
            setTimeout(status, 3000);
        },
        error: function (_jqXHR, textStatus, _errorThrown) {
            alert("Error: " + textStatus);
        }
    });
    alert("SSO Connect Web Server Restarted.");
}

/**
 * This pings the SSO Connect server to see whether it is alive.
 */
function simplePing() {
    return new Promise(function (resolve, reject) {
        $.ajax({
            url: "/ping/",
            async: true,
            type: 'GET',
            contentType: "text/plain; charset=UTF-8",
            success: function (response, _textStatus, _jqXHR) {
                if (response.status === "Ready") {
                    resolve(true);
                } else {
                    console.log("Server is not ready yet.");
                    reject(false);
                }
            },
            error: function (_jqXHR, _textStatus, _errorThrown) {
                console.log("Unable to contact server.");
                reject(false);
            }
        });
    });
}

/**
 * This retrieves the instance properties from the SSO Connect server.
 */
function serverAlive() {
    return new Promise(function (resolve, reject) {
        $.ajax({
            url: "/properties/"
        })
            .done(function () {
                getInstanceProperties(status);
                resolve(true);
            })
            .fail(function () {
                reject(false);
            });
    });
}

/**
 * Pings the SSO Connect server every second until it responds.
 */
function waitForServer() {
    let timeout = 1000; // milliseconds
    setTimeout(function () {
        simplePing()
            .then(function () {
                // The server is alive, do something here.
                // console.log("Server is available.");
            })
            .catch(function () {
                waitForServer(); // Try again
            });
    }, timeout);
}

/**
 * Saves the modified settings to the SSO Connect server which may cause it to restart.
 */
function saveSettings() {
    let sslFile = document.getElementById("ssl_file");
    if (sslFile.files.length > 0 && !(document.getElementById("jks").checked || document.getElementById("pfx").checked)) {
        alert("Please select the SSL Key Store type.");
        return;
    }

    document.getElementById('save_bottom').style.pointerEvents = 'none';
    document.getElementById('save_top').style.pointerEvents = 'none';
    document.getElementById('cancel_bottom').style.pointerEvents = 'none';
    document.getElementById('cancel_top').style.pointerEvents = 'none';

    let data = new FormData();
    data.append("sso_connect_host", document.getElementById("sso_connect_host").value);
    data.append("sso_ssl_port", document.getElementById("sso_ssl_port").value);
    data.append("key_store_password", document.getElementById("key_store_password").value);
    data.append("key_password", document.getElementById("key_password").value);
    if (document.querySelector('input[name=key_store_type]:checked')) {
        data.append("key_store_type", document.querySelector('input[name=key_store_type]:checked').value);
    }
    // data.append("key_type", $("#key_type_form input[name=key_type]:checked").val());
    data.append("map_email", document.getElementById("map_email").value);
    data.append("map_first_name", document.getElementById("map_first_name").value);
    data.append("map_last_name", document.getElementById("map_last_name").value);
    data.append("sign_idp_traffic", document.getElementById("sign_idp_traffic").checked);
    data.append("private_ip", document.getElementById("private_ip").value);
    data.append("private_port", document.getElementById("private_port").value);
    data.append("idp_type", document.getElementById("idp_type").value);
    let samlFile = document.getElementById("saml_file");
    if (samlFile.files.length > 0) {
        data.append("saml_file", samlFile.files[0]);
    }

    let slo_preference = "post";
    let sso_preference = "post";
    if (document.getElementById("sloPreferRedirect").style.display === "inline") {
        slo_preference = "redirect";
    }
    if (document.getElementById("ssoPreferRedirect").style.display === "inline") {
        sso_preference = "redirect";
    }

    data.append("slo_endpoint_preference", slo_preference);
    data.append("sso_endpoint_preference", sso_preference);

    // SKS properties
    // Send the edited set.  We will persist it on our side if the server-side save succeeds.
    data.append("sks_properties", JSON.stringify(edited_sks_properties));
    // console.log("Enabled? " + getSKSstate());
    // console.log("Sending sks properties: " + JSON.stringify(edited_sks_properties, 4));

    if (sslFile.files.length > 0) {
        data.append("ssl_file", sslFile.files[0]);
    }

    $.ajax({
        url: "/save_settings",
        async: false, // This needs to become asynchronous at some point.  Synchronous AJAX is not recommended anymore
        // async: true, // The interface is not working correctly if I make this async.  No time to debug right now.
        type: 'POST',
        data: data,
        cache: false,
        dataType: 'json',
        processData: false,
        contentType: false,
        success: function (response, _textStatus, _jqXHR) {
            document.getElementById('save_bottom').style.pointerEvents = 'auto';
            document.getElementById('save_top').style.pointerEvents = 'auto';
            document.getElementById('cancel_bottom').style.pointerEvents = 'auto';
            document.getElementById('cancel_top').style.pointerEvents = 'auto';
            if (response.result === "fail") {
                alert("Save failed: " + (response.message != null ? response.message : "unknown error encountered"));
            } else {
                saveSKSProperties();
                document.getElementById("ssl_file").value = "";
                document.getElementById("saml_file").value = "";
                document.getElementById("ssl_file_label").innerHtml = 'Drop a PFX or JKS file here or <span class="link_color">Click to Upload</span>';
                document.getElementById("ssl_selected").style.display = 'none';
                document.getElementById("saml_file_label").textinnerHtml = 'Drop an XML file here or <span class="link_color">Click to Upload</span>';
                setTimeout(function () {
                    getInstanceProperties(status);
                }, 1500);
                if (response.result_code === "re_login") {
                    logout();
                }
            }
        },
        error: function (_jqXHR, textStatus, _errorThrown) {
            document.getElementById('save_bottom').style.pointerEvents = 'auto';
            document.getElementById('save_top').style.pointerEvents = 'auto';
            document.getElementById('cancel_bottom').style.pointerEvents = 'auto';
            document.getElementById('cancel_top').style.pointerEvents = 'auto';
            alert("Error: " + textStatus);
        }
    });
}

function removeSSLFile() {
    // Remove any file that was uploaded
    document.getElementById("ssl_file_form").reset();

    removeFileOnServer('ssl');
}

function removeSAMLFile() {
    // Remove any file that was uploaded
    document.getElementById("saml_file_form").reset();

    removeFileOnServer('saml');
}

/**
 * Make a call to the server to remove any stored SSL or SAML file
 *
 * @param fileType 'ssl' or 'saml'
 * @returns nothing
 */
function removeFileOnServer(fileType) {

    let params = {
        "file_type": fileType
    };

    $.post("/remove_file", params,
        // success function
        function (response, _textStatus, _jqXHR) {
            if (response.result === "fail") {
                alert("Error removing file: " + (response.message != null ? response.message : "unknown error encountered"));
            } else {
                alert(fileType.toUpperCase() + " file removed");
                setTimeout(waitForServer, 1500);
                if (response.result_code === "re_login") {
                    logout();
                }
            }
        },
        'json');
}

function twofactor() {
    twoFactorToken = document.getElementById("two_factor_code").value;
    finishTwoFactor(twoFactorToken);
}

/**
 * Sends the 2FA code to the server, which completes the login if successful.
 * Otherwise, prompt the user to try again.
 */
function finishTwoFactor(twoFactorToken) {
    $.ajax({
        url: "/validate_2fa",
        async: false,
        type: 'post',
        data: {
            code: twoFactorToken,
            channel: twoFactorChannel
        },
        dataType: 'json',
        success: function (_response, _textStatus, _jqXHR) {
            if (_response["result"] === "success") {
                let action = _response["action"];
                if (action === "logged_in") {
                    isLoggedIn = true;
                    localStorage.setItem("lastLogin", document.getElementById("username").value);
                    document.getElementById("password").value = '';
                    document.getElementById("two_factor_code").value = '';
                    document.getElementById("device_approval_code").value = '';
                    getInstanceProperties(status);
                } else if (action === "start_login") {
                    document.getElementById("device_approval_code").value = '';
                    document.getElementById("two_factor_code").value = '';
                    getInstanceProperties(login);
                } else {
                    alert("Unknown action from 2fa_validate: " + action);
                }
            } else {
                alert("The two-factor code did not work. Please try again.\n\n" + _response["message"]);
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            alert("Unable to login.  Please try again.\n\n" + _errorThrown);
        }
    });
}

/**
 * This runs when the user selects a device approval channel.
 */
function selectDeviceApprovalChannel() {
    let channelSelector = document.getElementById("device_approval_channel");
    let channel = deviceApprovalChannels[channelSelector.selectedIndex];

    $.ajax({
        url: "/select_device_approval_channel",
        async: false,
        type: 'post',
        data: {
            channel: channel
        },
        dataType: 'json',
        success: function (_response, _textStatus, _jqXHR) {
            if (_response["result"] === "success") {
                let action = _response["action"];

                if (action === "logged_in") {
                    isLoggedIn = true;
                    localStorage.setItem("lastLogin", document.getElementById("username").value);
                    document.getElementById("password").value = '';
                    document.getElementById("device_approval_code").value = '';
                    getInstanceProperties(status);
                } else if (action === "require_device_approval") {
                    deviceApprovalChannel = _response["channel"];
                    document.querySelector("login-body").style.display = 'none';
                    document.querySelector("two-factor-dialog").style.display = 'none';
                    document.querySelector("device-approval-channel-dialog").style.display = 'none';
                    document.querySelector("device-approval-dialog").style.display = 'block';
                    $("device-approval-channel-dialog > div > p").text(getMessageForTwoFactorChannel(deviceApprovalChannel));
                    document.getElementById("device_approval_code").value = '';
                    document.getElementById("device-approval-channel-error").style.display = 'none';
                } else {
                    alert("Unknown action from select_device_approval_channel: " + action);
                }
            } else {
                alert("The device approval method selection did not work. Please try again.\n\n" + _response["message"]);
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            alert("Unable to login.  Please try again.\n\n" + _errorThrown);
        }
    });
}

/**
 * This runs when the user tells us that device approval happened in the background, and they don't want to enter a code.
 */
function performDeviceApprovalByOther() {
    // Just call start_login again after waiting a second.
    setTimeout(login, 1000);
}

/**
 * This runs when the user tells us that 2FA happened in the background, and they don't want to enter a code.
 */
function performTwoFactorByOther() {
    // Just call start_login again after waiting a second.
    setTimeout(login, 1000);
}


/**
 * This runs when the user selects a 2fa method.
 */
function selectTwoFactorChannel(evt, subChannel) {
    let channelSelector = document.getElementById("two_factor_channel");
    let channel = twoFactorChannels[channelSelector.selectedIndex];

    $.ajax({
        url: "/select_2fa_channel",
        async: false,
        type: 'post',
        data: {
            channel: channel,
            sub_channel: subChannel  // This may be null
        },
        dataType: 'json',
        success: function (_response, _textStatus, _jqXHR) {
            if (_response["result"] === "success") {
                let action = _response["action"];
                twoFactorChannel = _response["channel"];

                if (action === "logged_in") {
                    isLoggedIn = true;
                    localStorage.setItem("lastLogin", document.getElementById("username").value);
                    document.getElementById("password").value = '';
                    document.getElementById("two_factor_code").value = '';
                    getInstanceProperties(status);
                } else if (action === "start_login") {
                    setTimeout(login, 500);
                } else if (action === "require_2fa_channel") {
                    twoFactorChannels = _response["channel"];
                    if (twoFactorChannels) {
                        let twoFactorChannelSelector = document.getElementById('two_factor_channel');
                        twoFactorChannelSelector.options.length = 0;
                        twoFactorChannels.forEach(function (item, index) {
                            twoFactorChannelSelector.options[twoFactorChannelSelector.options.length] = new Option(getDisplayNameForTwoFactorChannel(item), index);
                        });
                    }
                    document.querySelector("login-body").style.display = 'none';
                    document.querySelector("device-approval-channel-dialog").style.display = 'none';
                    document.querySelector("device-approval-dialog").style.display = 'none';
                    document.querySelector("two-factor-dialog").style.display = 'none';
                    document.querySelector("two-factor-channel-dialog").style.display = 'block';
                    $("two-factor-channel-dialog > div > p").text(getMessageForTwoFactorChannel(twoFactorChannel));
                    document.getElementById("two-factor-channel-error").style.display = 'none';
                } else if ((action === "require_2fa") || (action === "require_2fa_for_duo")) {
                    twoFactorChannelUid = _response["channel_uid"];
                    setup2faDialog(twoFactorChannel, action);
                } else if (action === "wait_for_2fa") {
                    setTimeout(waitFor2FA, 500);
                } else {
                    alert("Unknown action from select_2fa_channel: " + action);
                    setTimeout(login, 500);
                }
            } else {
                alert("The 2FA method selection did not work. Please try again.\n\n" + _response["message"]);
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            alert("Unable to login.  Please try again.\n\n" + _errorThrown);
        }
    });
}

let two_fa_state = "wait";

async function waitFor2FA() {
    // Ping the server once a second to see if the confirmation has been received.
    // The server can reply with yes, no, or wait.  Either way we try the login again.

    let countdown = WAIT_FOR_2FA_APPROVAL_SECONDS;
    two_fa_state = "wait";
    while (countdown > 0) {
        await sleep(1000);
        askServerFor2FAapproval();
        if (two_fa_state === "approved") {
            document.getElementById("duo_select_dialog").style.display = 'none';
            setTimeout(login, 500);
            return;
        } else if (two_fa_state === "rejected") {
            alert("2FA was rejected");
            document.getElementById("duo_select_dialog").style.display = 'none';
            setTimeout(login, 500);
            return;
        } else if (two_fa_state === "error") {
            document.getElementById("duo_select_dialog").style.display = 'none';
            setTimeout(login, 500);
            return;
        }
        // else the state is "wait", try again in one second
        countdown--;
    }
    alert("2FA authentication timed out");
    setTimeout(login, 500);
}

/**
 * Use this as "await sleep(1000)".
 */
async function sleep(ms) {
    return new Promise(resolve => setTimeout(resolve, ms));
}

/**
 * Sets two_fa_state to "approved", "rejected", "wait", or "error".
 */
function askServerFor2FAapproval() {
    $.ajax({
        url: "/confirm_2fa",
        async: false,
        type: 'post',
        data: {},
        dataType: 'json',
        success: function (_response, _textStatus, _jqXHR) {
            if (_response["result"] === "success") {
                two_fa_state = _response["action"]; // "approved", "rejected", "wait"
            } else {
                alert("Failure while checking for 2FA approval.\n\n" + _response["message"]);
                return "error";
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            alert("Unable to login.  Please try again.\n\n" + _errorThrown);
            return "error";
        }
    });

}

function deviceApproval() {
    let deviceApprovalToken = document.getElementById("device_approval_code").value;
    finishDeviceApproval(deviceApprovalToken);
}

/**
 * Sends the device approval code to the server, which completes the login if successful.
 * Otherwise, prompt the user to try again.
 */
function finishDeviceApproval(deviceApprovalToken) {
    let twoFactorChannelUid;

    $.ajax({
        url: "/validate_device_verification_code",
        async: false,
        type: 'post',
        data: {
            code: deviceApprovalToken,
            channel: deviceApprovalChannel
        },
        dataType: 'json',
        success: function (_response, _textStatus, _jqXHR) {
            if (_response["result"] === "success") {
                let action = _response["action"];

                if (action === "logged_in") {
                    isLoggedIn = true;
                    localStorage.setItem("lastLogin", document.getElementById("username").value);
                    document.getElementById("password").value = '';
                    document.getElementById("device_approval_code").value = '';
                    getInstanceProperties(status);
                } else if (action === "start_login") {
                    setTimeout(login, 500);
                } else if (action === "require_2fa_channel") {
                    twoFactorChannels = _response["channel"];
                    if (twoFactorChannels) {
                        let twoFactorChannelSelector = document.getElementById('two_factor_channel');
                        twoFactorChannelSelector.options.length = 0;
                        twoFactorChannels.forEach(function (item, index) {
                            twoFactorChannelSelector.options[twoFactorChannelSelector.options.length] = new Option(getDisplayNameForTwoFactorChannel(item), index);
                        });
                    }
                    document.querySelector("login-body").style.display = 'none';
                    document.querySelector("device-approval-channel-dialog").style.display = 'none';
                    document.querySelector("device-approval-dialog").style.display = 'none';
                    document.querySelector("two-factor-dialog").style.display = 'none';
                    document.querySelector("two-factor-channel-dialog").style.display = 'block';
                    $("two-factor-channel-dialog > div > p").text(getMessageForTwoFactorChannel(twoFactorChannel));
                    document.getElementById("two-factor-channel-error").style.display = 'none';
                } else if (action === "require_2fa") {
                    twoFactorChannelUid = _response["channel_uid"];
                    setup2faDialog("email");
                } else {
                    alert("Unknown action from validate_device_verification_code: " + action);
                }
            } else {
                alert("The device approval code did not work. Please try again.\n\n" + _response["message"]);
            }
        },
        error: function (_jqXHR, _textStatus, _errorThrown) {
            alert("Unable to login.  Please try again.\n\n" + _errorThrown);
        }
    });
}


function enterKey(e, element) {
    if ((e.which || e.keyCode) === 13) {
        document.getElementById(element).click();
    }
}

$("#add_sks_property").on('click', addSKSProperty);
$("#clear_sks_properties").on('click', clearSKSProperties);
$("#config_sks_type").on('change', changeSKSType);
document.getElementById("login_button").addEventListener('click', login);
document.getElementById("initialize_button").addEventListener('click', initialize);
document.getElementById("device_approval_button").addEventListener('click', deviceApproval);
document.getElementById("device_approval_channel_button").addEventListener('click', selectDeviceApprovalChannel);
document.getElementById("device_approval_by_other_button").addEventListener('click', performDeviceApprovalByOther);
document.getElementById("two_factor_button").addEventListener('click', twofactor);
document.getElementById("two_factor_channel_button").addEventListener('click', selectTwoFactorChannel);
document.getElementById("2fa_by_other_button").addEventListener('click', performTwoFactorByOther);
document.getElementById("menu_item_status").addEventListener('click', function (_e) {
    getInstanceProperties(status);
});
document.getElementById("menu_item_config").addEventListener('click', function (_e) {
    getInstanceProperties(editConfiguration);
});
document.getElementById("menu_item_export").addEventListener('click', exportMetaData);
document.getElementById("cancel_bottom").addEventListener('click', status);
document.getElementById("cancel_top").addEventListener('click', status);
document.getElementById("save_bottom").addEventListener('click', saveSettings);
document.getElementById("save_top").addEventListener('click', saveSettings);
document.getElementById("menu_restart").addEventListener('click', restart);
document.getElementById("menu_item_logout").addEventListener('click', function (_e) {
    logout(false);
});
document.getElementById("password").addEventListener('keypress', function (e) {
    enterKey(e, "login_button");
});
document.getElementById("reset_sks_properties").addEventListener('click', resetSKSProperties);
document.getElementById("removeSSLCertificateFile").addEventListener('click', removeSSLFile);
document.getElementById("removeSamlMetadataFile").addEventListener('click', removeSAMLFile);
document.getElementById("two_factor_code").addEventListener('keypress', function (e) {
    enterKey(e, "two_factor_button");
});
document.getElementById("device_approval_code").addEventListener('keypress', function (e) {
    enterKey(e, "device_approval_button");
});
document.getElementById("duo_select_push").addEventListener('click', function (e) {
    select2faDuo("push");
});
document.getElementById("duo_select_sms").addEventListener('click', function (e) {
    select2faDuo("sms");
});
document.getElementById("duo_select_voice").addEventListener('click', function (e) {
    select2faDuo("voice");
});


document.getElementById("sks_config_running_checkbox").addEventListener("change",
    function (e) {
        if ($("#sks_config_running_checkbox").is(":checked")) {
            enableSKS(true);
        } else {
            enableSKS(false);
        }
    });

document.getElementById("saml_file").addEventListener('change', function (e) {
    let fileName = e.target.value.split('\\').pop();
    if (fileName) {
        document.getElementById("saml_file_label").innerHTML = sanitizeHTML(fileName);
    } else {
        document.getElementById("saml_file_label").innerHTML = 'Drop an XML file here or <span class="link_color">Click to Upload</span>';
    }
});
document.getElementById("ssl_file").addEventListener('change', function (e) {
    document.getElementById("keystore_pass_block").style.display = "none";
    document.getElementById("key_pass_block").style.display = "none";
    document.getElementById("key_type_block").style.display = "none";
    document.getElementById("jks").checked = false;
    document.getElementById("pfx").checked = false;
    let fileName = e.target.value.split('\\').pop();
    if (fileName) {
        document.getElementById("ssl_file_label").innerHTML = sanitizeHTML(fileName);
        document.getElementById("ssl_selected").style.display = 'block';
    } else {
        document.getElementById("ssl_file_label").innerHTML = 'Drop a PFX or JKS file here or <span class="link_color">Click to Upload</span>';
        document.getElementById("ssl_selected").style.display = 'none';
    }
});
document.getElementById("acs_value").addEventListener('click', function (e) {
    e.preventDefault();
});
document.getElementById("sls_value").addEventListener('click', function (e) {
    e.preventDefault();
});
document.getElementById("jks").addEventListener('click', function (_e) {
    document.getElementById("keystore_pass_block").style.display = "block";
    document.getElementById("key_pass_block").style.display = "block";
    // document.getElementById("key_type_block").style.display = "block";
});
document.getElementById("pfx").addEventListener('click', function (_e) {
    document.getElementById("keystore_pass_block").style.display = "block";
    document.getElementById("key_pass_block").style.display = "none";
    // document.getElementById("key_type_block").style.display = "block";
});

/**
 * Requests that the SSO Connect server perform a full sync.
 */
document.getElementById("menu_full_sync").addEventListener('click', function (_e) {
    $.ajax({
        url: "/full_sync",
        async: true,
        type: 'POST',
        contentType: "text/plain; charset=UTF-8",
        success: function (_response, _textStatus, _jqXHR) {
            // Wait for the server to respond before updating the screen
            setTimeout(status, 1000);
        },
        error: function (_jqXHR, textStatus, _errorThrown) {
            alert("Error: " + textStatus);
        }
    });
    alert("Full Sync Initiated.");
});

// KSC-269: Allow user to set preference for Post or Redirect
document.getElementById("sloChoosePost").addEventListener('click', function (ev) {
    setEndpointPreference('slo', 'post');
});
document.getElementById("ssoChoosePost").addEventListener('click', function (ev) {
    setEndpointPreference('sso', 'post');
});
document.getElementById("sloChooseRedirect").addEventListener('click', function (ev) {
    setEndpointPreference('slo', 'redirect');
});
document.getElementById("ssoChooseRedirect").addEventListener('click', function (ev) {
    setEndpointPreference('sso', 'redirect');
});


if (localStorage.lastLogin) {
    document.getElementById("username").value = localStorage.lastLogin;
}

getInstanceProperties();

// Start a timer to check on the status of the Keeperapp connection via the SSO Connect server
let keeperappTimer = setInterval(() => {
    checkKeeperappStatus()
}, SERVER_MONITOR_INTERVAL_SECS * 1000);

// You can call clearInterval(keeperappTimer) to stop the timer. 

/**
 * Sanitize and encode all HTML in a user-submitted string
 * https://portswigger.net/web-security/cross-site-scripting/preventing
 * @param  {String} str  The user-submitted string
 * @return {String} str  The sanitized string
 */
let sanitizeHTML = function (str) {
    return str.replace(/[^\w. ]/gi, function (c) {
        return '&#' + c.charCodeAt(0) + ';';
    });
};

// --------- Region Mappings from codes to names so that we don't have to update the software when we add a new region.
// --------- These are from the ISO-3166 tables.  The 2-letter codes are called "alpha-2" and the three-letter codes are called "alpha-3".

/**
 * Look up the region name for the region code, such as "United States" for "com" and "Australia" for "au".
 * If not found (unlikely) it returns the input
 */
function getRegionName(code) {

    code = code.toLowerCase();

    // Try the 2-letter code table first
    let regionName = regionMappings_2[code];
    if (regionName) {
        return regionName;
    }

    // Then the 3-letter code table
    regionName = regionMappings_3[code];
    if (regionName) {
        return regionName;
    }

    // If not found (unlikely) return the code itself
    return code;
}

let regionMappings_2 = {
    "eu": "Europe",
    "af": "Afghanistan",
    "ax": "Åland Islands",
    "al": "Albania",
    "dz": "Algeria",
    "as": "American Samoa",
    "ad": "Andorra",
    "ao": "Angola",
    "ai": "Anguilla",
    "aq": "Antarctica",
    "ag": "Antigua and Barbuda",
    "ar": "Argentina",
    "am": "Armenia",
    "aw": "Aruba",
    "au": "Australia",
    "at": "Austria",
    "az": "Azerbaijan",
    "bs": "Bahamas",
    "bh": "Bahrain",
    "bd": "Bangladesh",
    "bb": "Barbados",
    "by": "Belarus",
    "be": "Belgium",
    "bz": "Belize",
    "bj": "Benin",
    "bm": "Bermuda",
    "bt": "Bhutan",
    "bo": "Bolivia",
    "bes": "Bonaire, Sint Eustatius and Saba",
    "ba": "Bosnia and Herzegovina",
    "bw": "Botswana",
    "bv": "Bouvet Island",
    "br": "Brazil",
    "io": "British Indian Ocean Territory",
    "bn": "Brunei Darussalam",
    "bg": "Bulgaria",
    "bf": "Burkina Faso",
    "bi": "Burundi",
    "cv": "Cabo Verde",
    "kh": "Cambodia",
    "cm": "Cameroon",
    "ca": "Canada",
    "ky": "Cayman Islands",
    "cf": "Central African Republic",
    "td": "Chad",
    "cl": "Chile",
    "cn": "China",
    "cx": "Christmas Island",
    "cc": "Cocos (Keeling) Islands",
    "co": "Colombia",
    "cg": "Congo",
    "cd": "Congo, Democratic Republic",
    "ck": "Cook Islands",
    "cr": "Costa Rica",
    "ci": "Côte d'Ivoire",
    "hr": "Croatia",
    "cu": "Cuba",
    "cw": "Curaçao",
    "cy": "Cyprus",
    "cz": "Czech Republic",
    "dk": "Denmark",
    "dj": "Djibouti",
    "dm": "Dominica",
    "do": "Dominican Republic",
    "ec": "Ecuador",
    "eg": "Egypt",
    "sv": "El Salvador",
    "gq": "Equatorial Guinea",
    "er": "Eritrea",
    "ee": "Estonia",
    "et": "Ethiopia",
    "fk": "Falkland Islands (Malvinas)",
    "fo": "Faroe Islands",
    "fj": "Fiji",
    "fi": "Finland",
    "fr": "France",
    "gf": "French Guiana",
    "pf": "French Polynesia",
    "tf": "French Southern Territories",
    "ga": "Gabon",
    "gm": "Gambia",
    "ge": "Georgia",
    "de": "Germany",
    "gh": "Ghana",
    "gi": "Gibraltar",
    "gr": "Greece",
    "gl": "Greenland",
    "gd": "Grenada",
    "gp": "Guadeloupe",
    "gu": "Guam",
    "gt": "Guatemala",
    "gg": "Guernsey",
    "gn": "Guinea",
    "gw": "Guinea-Bissau",
    "gy": "Guyana",
    "ht": "Haiti",
    "hm": "Heard Island and McDonald Islands",
    "va": "Holy See (Vatican City State)",
    "hn": "Honduras",
    "hk": "Hong Kong",
    "hu": "Hungary",
    "is": "Iceland",
    "in": "India",
    "id": "Indonesia",
    "ir": "Iran, Islamic Republic Of",
    "iq": "Iraq",
    "ie": "Ireland",
    "im": "Isle of Man",
    "il": "Israel",
    "it": "Italy",
    "jm": "Jamaica",
    "jp": "Japan",
    "je": "Jersey",
    "jo": "Jordan",
    "kz": "Kazakhstan",
    "ke": "Kenya",
    "ki": "Kiribati",
    "kr": "Korea, Republic of",
    "kp": "North Korea (Democratic People's Republic of)",
    "kw": "Kuwait",
    "kg": "Kyrgyzstan",
    "la": "Lao People's Democratic Republic",
    "lv": "Latvia",
    "lb": "Lebanon",
    "ls": "Lesotho",
    "lr": "Liberia",
    "ly": "Libya",
    "li": "Liechtenstein",
    "lt": "Lithuania",
    "lu": "Luxembourg",
    "mo": "Macao",
    "mk": "North Macedonia",
    "mg": "Madagascar",
    "mw": "Malawi",
    "my": "Malaysia",
    "mv": "Maldives",
    "ml": "Mali",
    "mt": "Malta",
    "mh": "Marshall Islands",
    "mq": "Martinique",
    "mr": "Mauritania",
    "mu": "Mauritius",
    "yt": "The Department of Mayotte",
    "mx": "Mexico",
    "fm": "Micronesia, Federated States of",
    "md": "Republic of Moldova",
    "mc": "Monaco",
    "mn": "Mongolia",
    "me": "Montenegro",
    "ms": "Montserrat",
    "ma": "Morocco",
    "mz": "Mozambique",
    "mm": "Myanmar",
    "na": "Namibia",
    "nr": "Nauru",
    "np": "Nepal",
    "nl": "Netherlands",
    "nc": "New Caledonia",
    "nz": "New Zealand",
    "ni": "Nicaragua",
    "ne": "Niger",
    "ng": "Nigeria",
    "nu": "Niue",
    "nf": "Norfolk Island",
    "mp": "Northern Mariana Islands",
    "no": "Norway",
    "om": "Oman",
    "pk": "Pakistan",
    "pw": "Palau",
    "ps": "Palestine, State of",
    "pa": "Panama",
    "pg": "Papua New Guinea",
    "py": "Paraguay",
    "pe": "Peru",
    "ph": "Philippines",
    "pn": "Pitcairn",
    "pl": "Poland",
    "pt": "Portugal",
    "pr": "Puerto Rico",
    "qa": "Qatar",
    "re": "Réunion",
    "ro": "Romania",
    "ru": "Russian Federation",
    "rw": "Rwanda",
    "bl": "Saint Barthélemy",
    "sh": "Saint Helena, Ascension and Tristan da Cunha",
    "kn": "Saint Kitts and Nevis",
    "lc": "Saint Lucia",
    "mf": "Saint Martin (French part)",
    "pm": "Saint Pierre and Miquelon",
    "vc": "Saint Vincent and the Grenadines",
    "ws": "Samoa",
    "sm": "San Marino",
    "st": "Sao Tome and Principe",
    "sa": "Saudi Arabia",
    "sn": "Senegal",
    "rs": "Serbia",
    "sc": "Seychelles",
    "sl": "Sierra Leone",
    "sg": "Singapore",
    "sx": "Sint Maarten (Dutch part)",
    "sk": "Slovakia",
    "si": "Slovenia",
    "sb": "Solomon Islands",
    "so": "Somalia",
    "za": "South Africa",
    "gs": "South Georgia and the South Sandwich Islands",
    "es": "Spain",
    "lk": "Sri Lanka",
    "sd": "Sudan",
    "ss": "South Sudan",
    "sr": "Suriname",
    "sj": "Svalbard and Jan Mayen",
    "sz": "Swaziland",
    "se": "Sweden",
    "ch": "Switzerland",
    "sy": "Syrian Arab Republic",
    "tw": "Taiwan",
    "tj": "Tajikistan",
    "tz": "Tanzania, United Republic of",
    "th": "Thailand",
    "tl": "Timor-Leste",
    "tg": "Togo",
    "tk": "Tokelau",
    "to": "Tonga",
    "tt": "Trinidad and Tobago",
    "tn": "Tunisia",
    "tr": "Turkey",
    "tm": "Turkmenistan",
    "tc": "Turks and Caicos Islands",
    "tv": "Tuvalu",
    "ug": "Uganda",
    "ua": "Ukraine",
    "ae": "United Arab Emirates",
    "gb": "United Kingdom",
    "us": "United States of America",
    "um": "United States Minor Outlying Islands",
    "uy": "Uruguay",
    "uz": "Uzbekistan",
    "vu": "Vanuatu",
    "ve": "Venezuela",
    "vn": "Vietnam",
    "vg": "Virgin Islands, British",
    "vi": "Virgin Islands, U.S.",
    "wf": "Wallis and Futuna",
    "eh": "Western Sahara",
    // "yt": "Mayotte",
    "ye": "Yemen",
    "zm": "Zambia",
    "zw": "Zimbabwe",
}

let regionMappings_3 = {
    "com": "United States",
    "usgov": "US Government",
    "com.au": "Australia",
    "afg": "Afghanistan",
    "ala": "Åland Islands",
    "alb": "Albania",
    "dza": "Algeria",
    "asm": "American Samoa",
    "and": "Andorra",
    "ago": "Angola",
    "aia": "Anguilla",
    "ata": "Antarctica",
    "atg": "Antigua and Barbuda",
    "arg": "Argentina",
    "arm": "Armenia",
    "abw": "Aruba",
    "aus": "Australia",
    "aut": "Austria",
    "aze": "Azerbaijan",
    "bhs": "Bahamas",
    "bhr": "Bahrain",
    "bgd": "Bangladesh",
    "brb": "Barbados",
    "blr": "Belarus",
    "bel": "Belgium",
    "blz": "Belize",
    "ben": "Benin",
    "bmu": "Bermuda",
    "btn": "Bhutan",
    "bol": "Bolivia",
    "bes": "Bonaire, Sint Eustatius and Saba",
    "bih": "Bosnia and Herzegovina",
    "bwa": "Botswana",
    "bvt": "Bouvet Island",
    "bra": "Brazil",
    "iot": "British Indian Ocean Territory",
    "brn": "Brunei Darussalam",
    "bgr": "Bulgaria",
    "bfa": "Burkina Faso",
    "bdi": "Burundi",
    "cpv": "Cabo Verde",
    "khm": "Cambodia",
    "cmr": "Cameroon",
    "can": "Canada",
    "cym": "Cayman Islands",
    "caf": "Central African Republic",
    "tcd": "Chad",
    "chl": "Chile",
    "chn": "China",
    "cxr": "Christmas Island",
    "cck": "Cocos (Keeling) Islands",
    "col": "Colombia",
    "cog": "Congo",
    "cod": "Congo, Democratic Republic",
    "cok": "Cook Islands",
    "cri": "Costa Rica",
    "civ": "Côte d'Ivoire",
    "hrv": "Croatia",
    "cub": "Cuba",
    "cuw": "Curaçao",
    "cyp": "Cyprus",
    "cze": "Czech Republic",
    "dnk": "Denmark",
    "dji": "Djibouti",
    "dma": "Dominica",
    "dom": "Dominican Republic",
    "ecu": "Ecuador",
    "egy": "Egypt",
    "slv": "El Salvador",
    "gnq": "Equatorial Guinea",
    "eri": "Eritrea",
    "est": "Estonia",
    "swz": "Swaziland",
    "eth": "Ethiopia",
    "flk": "Falkland Islands (Malvinas)",
    "fro": "Faroe Islands",
    "fji": "Fiji",
    "fin": "Finland",
    "fra": "France",
    "guf": "French Guiana",
    "pyf": "French Polynesia",
    "atf": "French Southern Territories",
    "gab": "Gabon",
    "gmb": "Gambia",
    "geo": "Georgia",
    "deu": "Germany",
    "gha": "Ghana",
    "gib": "Gibraltar",
    "grc": "Greece",
    "grl": "Greenland",
    "grd": "Grenada",
    "glp": "Guadeloupe",
    "gum": "Guam",
    "gtm": "Guatemala",
    "ggy": "Guernsey",
    "gin": "Guinea",
    "gnb": "Guinea-Bissau",
    "guy": "Guyana",
    "hti": "Haiti",
    "hmd": "Heard Island and McDonald Islands",
    "vat": "Holy See (Vatican City State)",
    "hnd": "Honduras",
    "hkg": "Hong Kong",
    "hun": "Hungary",
    "isl": "Iceland",
    "ind": "India",
    "idn": "Indonesia",
    "irn": "Iran, Islamic Republic Of",
    "irq": "Iraq",
    "irl": "Ireland",
    "imn": "Isle of Man",
    "isr": "Israel",
    "ita": "Italy",
    "jam": "Jamaica",
    "jpn": "Japan",
    "jey": "Jersey",
    "jor": "Jordan",
    "kaz": "Kazakhstan",
    "ken": "Kenya",
    "kir": "Kiribati",
    "prk": "North Korea (Democratic People's Republic of)",
    "kor": "Korea, Republic of",
    "kwt": "Kuwait",
    "kgz": "Kyrgyzstan",
    "lao": "Lao People's Democratic Republic",
    "lva": "Latvia",
    "lbn": "Lebanon",
    "lso": "Lesotho",
    "lbr": "Liberia",
    "lby": "Libya",
    "lie": "Liechtenstein",
    "ltu": "Lithuania",
    "lux": "Luxembourg",
    "mac": "Macao",
    "mdg": "Madagascar",
    "mwi": "Malawi",
    "mys": "Malaysia",
    "mdv": "Maldives",
    "mli": "Mali",
    "mlt": "Malta",
    "mhl": "Marshall Islands",
    "mtq": "Martinique",
    "mrt": "Mauritania",
    "mus": "Mauritius",
    "myt": "The Department of Mayotte",  // "myt": "Mayotte",
    "mex": "Mexico",
    "fsm": "Micronesia, Federated States of",
    "mda": "Republic of Moldova",
    "mco": "Monaco",
    "mng": "Mongolia",
    "mne": "Montenegro",
    "msr": "Montserrat",
    "mar": "Morocco",
    "moz": "Mozambique",
    "mmr": "Myanmar",
    "nam": "Namibia",
    "nru": "Nauru",
    "npl": "Nepal",
    "nld": "Netherlands",
    "ncl": "New Caledonia",
    "nzl": "New Zealand",
    "nic": "Nicaragua",
    "ner": "Niger",
    "nga": "Nigeria",
    "niu": "Niue",
    "nfk": "Norfolk Island",
    "mkd": "North Macedonia",
    "mnp": "Northern Mariana Islands",
    "nor": "Norway",
    "omn": "Oman",
    "pak": "Pakistan",
    "plw": "Palau",
    "pse": "Palestine, State of",
    "pan": "Panama",
    "png": "Papua New Guinea",
    "pry": "Paraguay",
    "per": "Peru",
    "phl": "Philippines",
    "pcn": "Pitcairn",
    "pol": "Poland",
    "prt": "Portugal",
    "pri": "Puerto Rico",
    "qat": "Qatar",
    "reu": "Réunion",
    "rou": "Romania",
    "rus": "Russian Federation",
    "rwa": "Rwanda",
    "blm": "Saint Barthélemy",
    "shn": "Saint Helena, Ascension and Tristan da Cunha",
    "kna": "Saint Kitts and Nevis",
    "lca": "Saint Lucia",
    "maf": "Saint Martin (French part)",
    "spm": "Saint Pierre and Miquelon",
    "vct": "Saint Vincent and the Grenadines",
    "wsm": "Samoa",
    "smr": "San Marino",
    "stp": "Sao Tome and Principe",
    "sau": "Saudi Arabia",
    "sen": "Senegal",
    "srb": "Serbia",
    "syc": "Seychelles",
    "sle": "Sierra Leone",
    "sgp": "Singapore",
    "sxm": "Sint Maarten (Dutch part)",
    "svk": "Slovakia",
    "svn": "Slovenia",
    "slb": "Solomon Islands",
    "som": "Somalia",
    "zaf": "South Africa",
    "sgs": "South Georgia and the South Sandwich Islands",
    "ssd": "South Sudan",
    "esp": "Spain",
    "lka": "Sri Lanka",
    "sdn": "Sudan",
    "sur": "Suriname",
    "sjm": "Svalbard and Jan Mayen",
    "swe": "Sweden",
    "che": "Switzerland",
    "syr": "Syrian Arab Republic",
    "twn": "Taiwan",
    "tjk": "Tajikistan",
    "tza": "Tanzania, United Republic of",
    "tha": "Thailand",
    "tls": "Timor-Leste",
    "tgo": "Togo",
    "tkl": "Tokelau",
    "ton": "Tonga",
    "tto": "Trinidad and Tobago",
    "tun": "Tunisia",
    "tur": "Turkey",
    "tkm": "Turkmenistan",
    "tca": "Turks and Caicos Islands",
    "tuv": "Tuvalu",
    "uga": "Uganda",
    "ukr": "Ukraine",
    "are": "United Arab Emirates",
    "gbr": "United Kingdom",
    "usa": "United States of America",
    "umi": "United States Minor Outlying Islands",
    "ury": "Uruguay",
    "uzb": "Uzbekistan",
    "vut": "Vanuatu",
    "ven": "Venezuela",
    "vnm": "Vietnam",
    "vgb": "Virgin Islands, British",
    "vir": "Virgin Islands, U.S.",
    "wlf": "Wallis and Futuna",
    "esh": "Western Sahara",
    "yem": "Yemen",
    "zmb": "Zambia",
    "zwe": "Zimbabwe"
}
